(princ "We are now executing FILE2.LSP") (terpri)

(princ "Calling TEST-FUNCTION in FILE2.LSP") (terpri)

(test-function 1 1)

(princ "Finished") (terpri)

(princ "======================================================================") (terpri)
